# prepare

recommended operating systems:

* linux
* windows, install WSL Ubuntu: https://ubuntu.com/tutorials/install-ubuntu-on-wsl2-on-windows-10

required language and version: `nodejs 18.x`

run on the linux terminal...

```
# update system packages

sudo apt update

# install nodejs

sudo apt install nodejs
sudo apt install npm

# upgrade node to stable

sudo npm install n -g
sudo n stable
```

# install

in the base directory where the code resides...

```
# install nodejs libraries
npm install

# copy default configuration
cp .env.sample .env
```

# basic run

```
npm start
  OR
node server.js
```

# test/access/write robot brains

open a browser and access http://localhost:5001/

this is sufficient to debug robots via the browser-based code editor on the robot control panel

# developer/debug run

this is for developers working on the codebase, not robots!

```
npm install nodemon --global

nodemon
  OR
nodemon -L # WSL2, symlinked WIN folder (use --legacy-watch mode)
```

# TODO / notes / ideas

* TODO:
  * robot id and key verifier
  * league management
* IDEA: canvas reprojection - experimental rendering
* IDEA: rule cards
  * size/mass depends on code length

# load-balancing

this is optional - basic implementation that allows usage of multiple CPU cores

requires: nginx

example for system with 3 available cores

**1. configure `.env`**

specify number of cores in `PROCS`

```
PORT=5001
PROCS=3
```

**2. configure nginx**

server list must match in `upstream` and `map`

```
upstream web {
    least_conn;
    server 127.0.0.1:5001;
    server 127.0.0.1:5002;
    server 127.0.0.1:5003;
}

map $arg_balancerid $upserve {
    default "127.0.0.1:5001";
    2 "127.0.0.1:5002";
    3 "127.0.0.1:5003";
}

server {
    location ~* /socket.io {
        proxy_pass http://$upserve;

        proxy_http_version 1.1;
        proxy_set_header Upgrade $http_upgrade;
        proxy_set_header Connection "upgrade";
        proxy_set_header Host $http_host;

        proxy_set_header X-Real-IP $remote_addr;
        proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
        proxy_set_header X-Forwarded-Proto https;
        proxy_set_header X-Nginx-Proxy true;

        proxy_redirect off;
    }

    location / {
        proxy_pass http://web;

        proxy_http_version 1.1;
        proxy_set_header Upgrade $http_upgrade;
        proxy_set_header Connection "upgrade";
        proxy_set_header Host $http_host;

        proxy_set_header X-Real-IP $remote_addr;
        proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
        proxy_set_header X-Forwarded-Proto https;
        proxy_set_header X-Nginx-Proxy true;

        proxy_redirect off;
    }
}
```

**3. run in multi-server mode**

spawns multiple processes based on `PROCS`

```
node multiserver.js
```
