const brains = require("./battle/brains/cogitator.js");
const simulation = require("./battle/simulation.js");
const robotman = require("./robotman.js");

let io = false,
    Arenas = {};

const set_socket_io = (sock) => {
    io = sock;
}

const serversideArena = (listRobots, arenaOptions) => {
    let Arena = getArena(arenaOptions.arenaid);
    if (Arena) return Arena;

    let loadedBrains = [];

    let Robots = robotman.loadlist(listRobots);
    Object.values(Robots).forEach(R => loadedBrains.push(R));

    let numOfRobots = loadedBrains.length;

    if (numOfRobots <= 1)
        throw Error('[arenaman] at least two robots required');

    // XXX: config section via arenaOptions is messy
    // dev: requires refactor

    let arenaConfig = Object.assign({
        'maxRounds': 3,
        'autoRestart': 5,
        'seed': arenaOptions.arenaid,
    }, arenaOptions);

    let simConfig = {
        'sim': {
            'maxRounds': arenaConfig.maxRounds,
            'autoRestart': arenaConfig.autoRestart,
            'numOfRobots': numOfRobots,
            'defaultBrain': '/* do nothing */',
            'locked': true,
        },
        'brains': loadedBrains,
        'rules': {
            // 'arrangement': {
            //     'type': 'circular',
            //     'options': {
            //         'angleOffset': 3.14159
            //     }
            // },
            'limit': {
                'round': arenaOptions.limitRound || 60,
            }
        }
    };

    Arena = simulation.Factory(simConfig);

    Arena.set_simlog_id(arenaConfig.arenaid);

    Arena.set_socket_io(io);
    Arena.set_socket_room(arenaConfig.arenaid);

    Arenas[arenaConfig.arenaid] = Arena;

    return Arena;
}

const multipurposeArena = (arenaConfig) => {
    let Arena = getArena(arenaConfig.arenaid);
    if (Arena) return Arena;

    let minRobots = 2,
        maxRobots = 8,
        paramSim = {},
        paramBrains = [];

    paramSim["seed"] = arenaConfig.arenaid;

    if ('numOfRobots' in arenaConfig){
        // allow 2-8 robots
        paramSim["numOfRobots"] = parseInt(arenaConfig.numOfRobots) || minRobots
        paramSim["numOfRobots"] = Math.min(paramSim["numOfRobots"], maxRobots);
        paramSim["numOfRobots"] = Math.max(minRobots, paramSim["numOfRobots"]);
    }

    if ('manualBrains' in arenaConfig){
        let supplied = String(arenaConfig.manualBrains || '').trim(),
            numManuals = 0;

        if (!supplied) {
            // empty param, ignore
        } else if (isNaN(supplied)) {
            // could refer to a robot brain (or several, comma-separated)
            supplied = supplied.split(",");
            numManuals = supplied.length;
            /* load server-side brains
                dev: users may request -MANUAL- (socket) connectors, this
                is dependent on the "Debug X" naming convention.
                therefore, sort the list so manual appears on top and is
                processed first so that the connectors will acquire
                Debug 1, Debug 2, etc names */
            supplied.sort();
            supplied.forEach(robotid => {
                if (robotid == '-MANUAL-') {
                    paramBrains.push(
                        brains.driver.use_socket(
                            arenaConfig.clientConnection())
                    );
                } else {
                    var brain = robotman.load(robotid);
                    if (brain) {
                        console.log(
                            "[sockman] server-side robot",
                            arenaConfig.arenaid, robotid);
                        paramBrains.push(brain);
                    } else {
                        console.log(
                            "[sockman] server-side robot not found",
                            arenaConfig.arenaid, robotid);
                    }
                }
            });
        } else {
            // default: use socket connectors for remote test/training

            // sanity: prevent -ve number of manuals
            numManuals = Math.max(
                parseInt(arenaConfig.manualBrains) || 0,
                0);

            // sanity: keep within maxRobots
            numManuals = Math.min(
                numManuals,
                paramSim["numOfRobots"] || maxRobots);

            // load the socket connectors
            for (var i=0; i<numManuals; i++) {
                paramBrains.push(
                    brains.driver.use_socket(
                        arenaConfig.clientConnection())
                );
            }
        }
    }

    Arena = simulation.Factory({
        "sim": paramSim,
        "brains": paramBrains
    });

    Arena.set_simlog_id(arenaConfig.arenaid);

    Arena.set_socket_io(io);
    Arena.set_socket_room(arenaConfig.arenaid);

    Arenas[arenaConfig.arenaid] = Arena;

    return Arena;
}

const getArena = (arenaid) => {
    if (arenaid in Arenas) return Arenas[arenaid];
    return false;
}

const removeArena = (arenaid) => {
    if (arenaid in Arenas) {
        Arenas[arenaid].stop();
        delete Arenas[arenaid];
        console.log("[arenaman] removed arena: ", arenaid);
    }
}

module.exports = {
    set_socket_io: set_socket_io,
    Arena : {
        multipurpose: multipurposeArena,
        serverside: serversideArena,
    },
    Arenas: Arenas,
    get: getArena,
    remove: removeArena,
};
