const fs = require('fs');

const { resolve } = require("path");

const pathlogs = resolve("./static/simlogs/");

console.log("[simlog]", pathlogs);

const hashname = (simid) => {
    return simid;
};

const write = (simid, roundnum, robotnames, placements) => {
    var fullpath = resolve(pathlogs + "/" + hashname(simid) + ".csv");

    var columns = [
        "placement",
        "lifetime",
        "health",
        "accuracy"
    ];

    // acquire headers for structure checking
    var headers = [
        "Date-Time",
        "Round"
    ];
    robotnames.forEach(robotname => {
        columns.forEach(colname =>
            headers.push(robotname + " " + colname));
    });

    // generate text line for headers
    var row_headers = '"' + headers.join('","') + '"';

    var create_file = false;

    if (fs.existsSync(fullpath)) {
        var datalen = row_headers.length,
            buffer = Buffer.alloc(datalen),
            fh = fs.openSync(fullpath, 'r');
        fs.readSync(fh, buffer, 0,  datalen, 0);
        var file_headers = buffer.toString('utf8', 0, datalen);
        if (file_headers == row_headers) {
            console.log("[simlog] identical headers");
            create_file = false;
        } else {
            console.log(row_headers);
            console.log(file_headers);
            console.log("[simlog] non-matching headers");
            create_file = true;
        }
    } else {
        create_file = true;
    }

    if (create_file) {
        // generate new csv if it does not exist
        fs.closeSync(fs.openSync(fullpath, 'w'))

        fs.appendFileSync(fullpath, row_headers + "\n");

        console.log("[simlog]", fullpath, "created");
    }

    // get metadata and placement by robot name
    var by_robotname = {};
    placements.map((x, idx) => {
        x.participants.map(part => {
            by_robotname[part[0]] = {
                placement   : idx + 1,
                lifetime    : part[3],
                health      : part[6],
                accuracy    : part[7]
            }
        })
    });

    // generate flat results
    var row = [],
        now = new Date(),
        dt = now.getFullYear() + "-"
            + (("0" + (now.getMonth()+1)).slice(-2)) + "-"
            + (("0" + (now.getDate())).slice(-2)) + " "
            + (("0" + (now.getHours())).slice(-2)) + ":"
            + (("0" + (now.getMinutes())).slice(-2)) + ":"
            + (("0" + (now.getSeconds())).slice(-2));

    row.push(dt);
    row.push(roundnum)
    robotnames.forEach(robotname => {
        var metadata = by_robotname[robotname];
        columns.forEach(colname => row.push(metadata[colname]))
    });
    console.log("[simlog] written round", roundnum);

    fs.appendFileSync(fullpath, row.join(",") + "\n");
}

const write_json = (simid, roundnum, placements) => {
    var fullpath = resolve(pathlogs + "/" + hashname(simid) + ".json");

    let logs = [];
    if (fs.existsSync(fullpath)) {
        logs = JSON.parse(fs.readFileSync(fullpath));
    }

    var entry = {
        'datetime': new Date(),
        'round': roundnum,
        'placements': placements
    };

    logs.push(entry);

    fs.writeFileSync(fullpath, JSON.stringify(logs));
}

const read_json = (simid) => {
    var fullpath = resolve(pathlogs + "/" + hashname(simid) + ".json");

    let logs = false;
    if (fs.existsSync(fullpath)) {
        logs = JSON.parse(fs.readFileSync(fullpath));
    }

    return logs;
}

const check = (simid) => {
    var fullpath = resolve(pathlogs + "/" + hashname(simid) + ".csv");

    if (fs.existsSync(fullpath)) return fullpath;

    return false;
}

const remove = (simid) => {
    fullpath = check(simid);
    if (fullpath) {
        fs.unlinkSync(fullpath);
        console.log("[simlog] remove,", fullpath);
    } else {
        console.log("[simlog] remove, not found:", fullpath);
    }
}

module.exports = {
    write: write,
    write_csv: write,
    write_json: write_json,
    read_json: read_json,
    check: check,
    remove: remove,
    hashname: hashname
};
