let ArenaVectorBodies = (context, fontfamily, fontsize) => {
    var ctx = context,
        fontfamily = fontfamily || "sans-serif",
        fontsize = fontsize || 12
        fontdefault = fontsize + "px " + fontfamily;

    const reproject = (x, y) => {
        // future: reprojection for more creative rendering of vectors
        return [x, y];
    }

    const draw_generic_vertexes = (body) => {
        ctx.beginPath();
        body.edges.forEach((e) => ctx.lineTo(...reproject(e.x, e.y)));
        ctx.closePath();
        ctx.strokeStyle = "#000"; // default
        ctx.fillStyle = body.fillStyle;
        ctx.fill();
    };

    const draw_robot = (body) => {
        ctx.beginPath();
        body.edges.forEach((e) => ctx.lineTo(...reproject(e.x, e.y)));
        ctx.closePath();

        ctx.lineWidth = 3;
        if (body.active) {
            // ctx.fillStyle = body.fillStyle;
            ctx.strokeStyle = body.fillStyle;
        } else {
            // ctx.fillStyle = "#000";
            ctx.strokeStyle = "transparent";
            ctx.strokeStyle = "black";
        }
        ctx.stroke();
        // ctx.fill();

        ctx.textAlign = "center";
        ctx.textBaseline = "top";
        ctx.font = fontdefault;

        if (body.active) ctx.fillStyle = "white";
        else ctx.fillStyle = "red";

        ctx.fillText(
            body.entityName,
            ...reproject(
                body.center[0],
                body.center[1] - body.maxDimension));
    };

    const draw_sensor = (body) => {
        // 0 (tr), 3 = farthest edge; 1 (tl), 2 = closest edge

        let // find offsets from farthest edge center
            ox = (body.edges[0].x - body.edges[3].x) / 2,
            oy = (body.edges[0].y - body.edges[3].y) / 2,
            // calculate nearest edge point
            // shape is trapezoid, so imaginary rectangle tl edge
            x1 = (body.edges[1].x + body.edges[2].x) / 2 - ox,
            y1 = (body.edges[1].y + body.edges[2].y) / 2 - oy,
            // angle - rotated to be perpendicular
            ang = body.angle + 1.5708;

        let len = body.scanDistance,
            grad = ctx.createLinearGradient(
                x1, y1, x1 + Math.cos(ang) * len,
                        y1 + Math.sin(ang) * len);

        grad.addColorStop(0, "rgb(150,150,255,0)");
        grad.addColorStop(0.05, "rgb(150,150,255,0.15)");
        grad.addColorStop(0.8, "rgb(150,150,255,0.15)");
        grad.addColorStop(1, "rgba(150,150,255,0)");

        ctx.beginPath();
        body.edges.forEach(e => ctx.lineTo(...reproject(e.x, e.y)));
        ctx.fillStyle = grad
        ctx.closePath();
        ctx.fill();

        // ctx.strokeStyle="white"; ctx.stroke();
    };

    const draw_projectiles = (body) => {
        ctx.beginPath();
        ctx.arc(
            ...reproject(body.center[0], body.center[1]),
            body.radius, 0, 2 * Math.PI, false);
        ctx.closePath();

        ctx.fillStyle = "#fff";
        ctx.fill();
    };

    return {
        draw: (robots, walls, projectiles, sensors) => {
            walls.forEach(wall => draw_generic_vertexes(wall, ctx));
            sensors.forEach(sensor => draw_sensor(sensor, ctx));
            robots.forEach(robot =>  draw_robot(robot, ctx));
            projectiles.forEach(projectile => draw_projectiles(projectile, ctx));
        }
    };
}
