function createArenaVisuals(context, fontfamily, fontsize) {
// factory to ensure local scoping of persistent variables

let State = {
    "spritesLoaded": false,
    "explodingRobots": []
};

let EntitySprites = Animator({
    "explosion_continuous":
        ["/artwork/explosion.512x512.16.png",
            // (debug) play continuously - right to left
            Animator.Factory.spriteCurrentFrame (
                128, 128, 12, (f, fc) => {
                if (!f) f = fc;
                f--;
                return f;
            })
        ],
    "explosion_oneoff": {
        "src": "/artwork/explosion.512x512.16.png",
        "before": [
            // play once only - left to right
            Animator.Factory.spriteCurrentFrame(
                128, 128, 1, (f, fc, meta) => {
                if (typeof f == "undefined") f = -1;
                f++;
                return f;
            })
        ],
        "after": [
            // cleanup robots after completely exploded
            function(src, tgt, meta){
                if (meta.norender
                        && meta.srcid
                        && State.explodingRobots.includes(meta.srcid)){
                    State.explodingRobots.splice(
                        State.explodingRobots.indexOf(meta.srcid), 1);
                }
            }
        ]
    },
    "robot":
        ["/artwork/robot.48x30.2.png",
            // switch frames based on state of robot
            Animator.Factory.spriteCurrentFrame(
                24, 30, 1,
                (f, fc, ctick, meta) => meta.active ? 0:1
            )
        ],
    "floor": "/artwork/plate.64x64.png"
}, (num) => {
    State.spritesLoaded = true;
});

return ((context, fontfamily, fontsize) => {
    var ctx = context,
        fontfamily = fontfamily || "sans-serif",
        fontsize = fontsize || 12
        fontdefault = fontsize + "px " + fontfamily,
        currentFrame = false,
        robotById = false,
        robot_count = 0,
        font_datauris = [],
        fonts_loaded = false;

    (async function() {
        // retrieve custom fonts from document css styles
        var fonts = [];
        for (const ss of document.styleSheets) {
            try {
                // external style sheet rules cannot be accessed
                // https://stackoverflow.com/a/49160760
                var rules = ss.cssRules;
            } catch(e) {
                console.log("cannot read rules from", ss.href);
                continue;
            }
            for (const rule of rules) {
                if (rule.cssText.startsWith("@font-face")) {
                    var fontFamily = rule.style.fontFamily,
                        url = rule.style.src.match(/url\(\"(.*?)\"\)/);
                    if (url) fonts.push([fontFamily, url[1]]);
                }
            };
        };
        // preload custom fonts, store as datas uri
        for (const font of fonts) {
            let blob = await fetch(font[1]).then(r => r.blob());
            let datauri = await new Promise(resolve => {
                let reader = new FileReader();
                reader.onload = () => resolve(reader.result);
                reader.readAsDataURL(blob);
            });
            console.log("font preloaded:", font[0], datauri.length);
            font_datauris.push([font[0], datauri]);
        }
        fonts_loaded = true; // completed
    })();

    const sprite_robot_body = (id) => {
        var body = robotById[id];
        EntitySprites.to_canvas(
            ctx, "robot",
            body.center[0], body.center[1],
            body.entityWidth, body.entityHeight, body.angle,
            { tick: currentFrame,
              id: "r." + body.id,
              active: body.active }
        );
    };

    const sprite_robot_explosion = (id) => {
        var body = robotById[id];
        EntitySprites.to_canvas(
            ctx, "explosion_oneoff",
            body.center[0], body.center[1],
            body.maxDimension * 3, body.maxDimension * 3, body.angle,
            { tick: currentFrame,
              id: "r.e." + id,
              srcid: id }
        );
    };

    var layers = {};

    var styles_shared = '' +
        '<style>' +
            'table#ranks { ' +
                'font-size:12px;' +
                'border-spacing: 2px 1px;' +
                'font-family: sans-serif; }' +
            'table#ranks tr { ' +
                'color: rgba(255,255,255,0.6); ' +
                'background: rgba(100,100,100,0.6); }' +
            //'table#ranks tr:nth-child(2n+1) { ' +
            //    'background: rgba(100,100,100,0.6); }' +
            'table#ranks tr.pos-1 { ' +
                'color: black; ' +
                'background: linear-gradient(to right, #f8b500, #fceabb); }' +
            'table#ranks tr.pos-2 { ' +
                'color: black; ' +
                'background: linear-gradient(to right, #757f9a, #d7dde8); }' +
            'table#ranks tr.pos-3 { ' +
                'color: white; ' +
                'background: linear-gradient(to right, #603813, #b29f94); }' +
            'table#ranks td, table#ranks th { ' +
                'vertical-align: middle; ' +
                'text-align: center; ' +
                'padding: 2px; }' +
            'table#ranks th { ' +
                'color: white; ' +
                'padding: 5px; ' +
                'background: rgba(0,51,102,0.7); }' +
            'table#ranks .pos { ' +
                'font-size: 15px; ' +
                'font-weight: bold; }' +
            'table#ranks td div { ' +
                'margin-bottom: 2px; ' +
                'display: flex; ' +
                // 'border: 1px solid red; ' +
                'justify-content: space-between; }' +
            'table#ranks td div:last-child { ' +
                'margin-bottom: 0; }' +
            'table#ranks .b { ' +
                'font-size: 9px; ' +
                'padding: 2px 4px 1px 4px; ' +
                'border-radius: 3px 3px; ' +
                'color: white; ' +
                'background: purple; }' +
        '</style>' ;

    const layer_html = (layerid, html, options) => {
        // render fixed-width html
        // supports (ratecapped) content update, custom fonts, dynamic height

        if (!fonts_loaded) return false;

        let NOW = (new Date).getTime() / 1000,
            cw = ctx.canvas.width,
            ch = ctx.canvas.height;

        options = Object.assign({
            width: 300,
            height: false, // auto-detected
            xmult: 0.5, // horizontally-centred
            ymult: 0.5, // vertically-centred
            ratecap: 0.5 // every 200ms
        }, options);

        var layer = false;
        if (!(layerid in layers)) {
            layer = {
                "updated": NOW,
                "html": false,
                "final": document.createElement('canvas'),
                "image": new Image()
            };
            layers[layerid] = layer;
        } else {
            layer = layers[layerid];
        }

        if (NOW - layer.updated >= options.ratecap && html != layer.html) {
            // refresh required
            if (!options.height) {
                // generate hidden html element to estimate height
                let divcalc = document.createElement("div"),
                    container = document.createElement("div");
                divcalc.style.height = "0px";
                divcalc.style.width = "0px";
                divcalc.style.overflow = "hidden";
                divcalc.append(container);
                // container behaves as a pseudo-page
                // height expands to html
                container.style.width = options.width + "px";
                container.innerHTML = html;
                document.body.appendChild(divcalc);
                options.height = container.offsetHeight;
                // cleanup
                document.body.removeChild(divcalc);
            }

            let fontfaces = font_datauris.map(x =>
                "@font-face { "
                    + "font-family: " + x[0] + "; "
                    + "src: url('" + x[1] + "');  }").join(" ");

            var data =
            '<svg xmlns="http://www.w3.org/2000/svg" ' +
              'width="' + options.width + '" ' +
              'height="' + options.height + '">' +
                '<foreignObject width="100%" height="100%">' +
                    '<div xmlns="http://www.w3.org/1999/xhtml">' +
                    '<style>' + fontfaces + '</style>' +
                    html +
                    '</div>' +
                '</foreignObject>' +
            '</svg>',
            DOMURL = self.URL || self.webkitURL || self,
            svg = new Blob([data], {type: "image/svg+xml;charset=utf-8"}),
            url = DOMURL.createObjectURL(svg);

            layer.image.onload = function(){
                DOMURL.revokeObjectURL(url);
                // double-buffer write to prevent flickering
                layer.final.width = layer.image.width;
                layer.final.height = layer.image.height;
                layer.final.getContext("2d").drawImage(layer.image, 0, 0);
                layer.updated = NOW;
            };

            layer.html = html;
            layer.image.src = url; // trigger load/reload
        }

        if (layer.final) {
            ctx.drawImage(
                layer.final,
                (cw - layer.final.width) * options.xmult,
                (ch - layer.final.height) * options.ymult);
        }
    };

    const html_wrap = (tag, obj) => {
        // wraps obj in specified tag with attributes
        // obj can be primitive or object {value: ..., [attributes]}
        // returns list [<start tag>, content, <end tag>]
        var attrs = [],
            html = obj;
        if (typeof(obj) === "object" && !Array.isArray(obj)) {
            for (span of ["rowspan", "colspan"]) {
                if (span in obj) {
                    var val = obj[span];
                    if (val === false) continue
                    else if (val > 0) attrs.push(span + '="' + val + '"')
                    else if (val <= 0) return false;
                }
            };
            for (attr of ["class", "style"]) {
                if (attr in obj)
                    attrs.push(attr + '="' + obj[attr] + '"');
            }
            html = obj.value;
        }
        attrs = attrs.join(" ");
        if (attrs) attrs = " " + attrs;
        return ["<" + tag + attrs + ">", html, "</" + tag + ">"];
    };

    const html_row_cells = (rows, options) => {
        // return html table rows and cells with attributes
        // expects rows to be list (rows) of lists (cols)
        // column value can be primitive or object (re: html_wrap)
        options = Object.assign({
            tag: "td",
            cell: (rowidx, colidx, val) =>
                (Array.isArray(val) ?
                    "[" + rowidx + ", "  + colidx + ", " + val[0] + "] " +
                        "(" + val.length + ")"
                    : val)
        }, options);

        if (Array.isArray(rows) && rows.every(i => (typeof i === "string")))
            rows = [rows];

        var html = [];
        for (var rowidx=0; rowidx < rows.length; rowidx++){
            var rowdef = html_wrap("tr", rows[rowidx]);
                cols = rowdef[1],
                row = [];
            for (var colidx=0; colidx < cols.length; colidx++){
                tokens = html_wrap(options.tag, cols[colidx]);
                if (tokens === false) continue; // don't render
                tokens[1] = options.cell(rowidx, colidx, tokens[1]);
                row.push(tokens.join(""));
            }
            rowdef[1] = row.join("");
            html.push(rowdef.join(""));
        }

        return html.join("");
    };

    let results_cache = {};

    const process_results = (results) => {
        // server already "placed" (1,2...) robots/teams
        // generation only runs once per new results (i.e. per round)

        if (!results_cache.robots) {
            var teamcounts = [];
            results_cache.robots = results.robots.reduce((acc, x, idx)=>{
                x.participants.forEach((r, ridx) => {
                    let rowspan = false;
                    if (x.participants.length > 1) {
                        if (ridx > 0) rowspan = -x.participants.length;
                            else rowspan = x.participants.length;
                    }
                    teamcounts[r[1]] = (teamcounts[r[1]] || 0) + 1;
                    var pos = idx + 1;
                    acc.push({
                        class: "pos-" + pos,
                        value: [
                            {
                                value: pos,
                                class: "pos",
                                rowspan: rowspan
                            },
                            r[0],
                            r[3].toFixed(4),
                            r[6],
                            r[7].toFixed(2) + "% (" + r[4] + "/" + r[5] + ")",
                        ]}
                    );
                });
                return acc;
            }, []);

            // generate additional metadata and express them via flags
            // these are useful to determine what to render later

            for (const [team, count] of Object.entries(teamcounts)) {
                // MultiRobotTeams = more than 1 robot per teams
                if (count > 1) {
                    results_cache["MultiRobotTeams"] = true;
                }
            }

            if (Object.keys(teamcounts).length == 1) {
                // SingleTeamOnly
                results_cache["SingleTeamOnly"] = true;
            }
        }

        if (!results_cache.teams) {
            results_cache.teams = results.teams.reduce((acc, x, idx)=>{
                x.participants.forEach((t, ridx) => {
                    let rowspan = false;
                    if (x.participants.length > 1) {
                        if (ridx > 0) rowspan = -x.participants.length;
                            else rowspan = x.participants.length;
                    }
                    var pos = idx + 1;
                    acc.push({
                        class: "pos-" + pos,
                        value: [
                            {
                                value: pos,
                                class: "pos",
                                rowspan: rowspan
                            },
                            t[0],
                            (t[1].toFixed(2) == t[2].toFixed(2) ?
                                t[1].toFixed(2)
                                : t[1].toFixed(2) + " / " + t[2].toFixed(2)),
                            t[4],
                            t[5].toFixed(2)
                        ]}
                    );
                });
                return acc;
            }, []);
        }
    }

    const draw_results = () => {
        let cw = ctx.canvas.width;

        var render = ["robots"];

        if (results_cache.MultiRobotTeams && !results_cache.SingleTeamOnly) {
            // only render teams when its possible to distinguish teams and robots
            render.push("teams");
        }

        if (render.includes("robots")) {
            layer_html("results-robots",
                styles_shared +
                '<table id="ranks" width="100%">' +
                    '<tr><th colspan="6">' +
                        'Round Results: Robots' +
                    '</th></tr>' +
                    html_row_cells(
                        ["Place", "Robot",
                            "Seconds",
                            "Health",
                            "Accuracy"],
                        {"tag": "th"}) +
                    html_row_cells(results_cache.robots) +
                '</table>', {
                    "width":
                        Math.min(
                            cw/render.length,
                            cw/1.2) + 1,
                    "xmult":
                        render.length == 2 ? 0 : 0.5
                });
            // end: layer_html(results-robots)
        }

        if (render.includes("teams")) {
            layer_html("results-teams",
                styles_shared +
                '<table id="ranks" width="100%">' +
                    '<tr><th colspan="5">' +
                        'Round Results: Teams' +
                    '</th></tr>' +
                    html_row_cells(
                        ["Place", "Teams",
                            "Min / Max (s)",
                            "Avg Health",
                            "Avg Accuracy"],
                        {"tag": "th"}) +
                    html_row_cells(results_cache.teams) +
                '</table>', {
                    "width":
                        Math.min(
                            cw/render.length,
                            cw/1.2) + 1,
                    "xmult":
                        render.length == 2 ? 1 : 0.5
                });
            // end: layer_html(results-team)
        }
    }

    return {
        update: (frame, robots) => {
            currentFrame = frame;
            robotById = robots;
            robot_count = Object.keys(robots).length;
            return State;
        },
        reset: () => {
            Animator.Factory.clear_saved_states();
            State.explodingRobots = [];
            currentFrame = false;
            robotById = {};
            robot_count = 0;
        },
        background: () => {
            if (!State.spritesLoaded) return;
            let cw = ctx.canvas.width,
                ch = ctx.canvas.height;
            EntitySprites.as_pattern
                (ctx, "floor",
                cw, ch,
                ((cw % 64) - 64) / 2,
                ((ch % 64) - 64) / 2);
        },
        render: () => {
            if (currentFrame === false) return;
            if (!State.spritesLoaded) return;
            if (!robot_count) return;

            // turn off smoothing for pixel art
            // preserve state for restore later
            let _smoothingEnabled = ctx.imageSmoothingEnabled;
            ctx.imageSmoothingEnabled = false;

            Object.values(robotById).forEach(robot => {
                sprite_robot_body(robot.id);
            });

            State.explodingRobots =
                State.explodingRobots.filter(id => id in robotById);

            State.explodingRobots.forEach(id => {
                sprite_robot_explosion(id);
            });

            ctx.imageSmoothingEnabled = _smoothingEnabled;
        },
        explode_robot: (id) => {
            if (State.explodingRobots.includes(id)) return;
            State.explodingRobots.push(id);
        },
        hud: (state) => {
            let cw = ctx.canvas.width,
                ch = ctx.canvas.height;

            ctx.setTransform(1, 0, 0, 1, 0, 0);

            ctx.fillStyle = "#fff";
            ctx.font = fontdefault;
            ctx.textBaseline = 'top';
            ctx.textAlign = 'left';
            var status =
                "round: " + state.count.rounds + " " +
                "frame: " + state.count.frames + " " +
                "secs: " + state.count.seconds.toFixed(2);
            ctx.fillText(status, 10, 10);

            if (state.round.remaining <= 10) {
                ctx.font = (fontsize * 2) + "px " + fontfamily;
                ctx.textBaseline = 'alphabetic';
                ctx.textAlign = 'right';
                ctx.fillStyle = "#fff";
                ctx.fillText(
                    "ending: " + state.round.remaining.toFixed(2),
                    cw - 10, ch - 10);
            };
        },
        results: (state) => {
            if (
                !('round' in state) ||
                !('results' in state.round) ||
                !state.round.results
            ) {
                // clears cache on beginning of new round
                results_cache = {};
                return;
            }

            process_results(state.round.results);
            // src: simulation.js::determine_round_results

            draw_results();
        }
    }
})(context, fontfamily, fontsize);

} // createArenaVisuals()
