const createServerMonitor = (container, css) => {
    let apikey = false,
        list_lb = document.querySelector(container),
        last_report = (new Date()).getTime();

    if (!css) css = {};

    if (!css.server) css.server = "f6 ph2 pv1 mb2 mr1";
    if (!css.error) css.error = "b red";
    if (!css.nodata) css.nodata = "b yellow";

    if (!css.ago_nominal) css.ago_nominal = "black bg-green";
    if (!css.ago_warning) css.ago_warning = "black bg-yellow";
    if (!css.ago_danger) css.ago_danger = "black bg-red";

    const setAPIKey = (key) => {
        apikey = key;
        console.log("[ui.serverMonitor.js] APIKey set", apikey.length, "chars");
    };

    let onData = (data, last_tracked) => {};
    let onError = (err, last_tracked) => {};

    const setOnData = (func) => onData = func;
    const setOnError = (func) => onData = func;

    const update = () => {
        let acquireData = false;

        if (apikey) {
            acquireData = (url) => {
                return fetchTimeout(url, {
                    method: 'POST',
                    mode: 'cors',
                    cache: 'no-cache',
                    headers: {
                        'Content-Type': 'application/json'
                    },
                    body: JSON.stringify({ 'api': apikey }),
                });
            };
        } else {
            acquireData = (url) => {
                return fetchTimeout(url, {
                    mode: 'cors',
                    cache: 'no-cache',
                    headers: {
                        'Content-Type': 'application/json'
                    },
                });
            };
        }

        acquireData('/api/status')
            .then(res => res.json())
            .then(data => {
                let last_tracked = Math.round((((new Date()).getTime() - last_report)/1000));

                onData(data, last_tracked);

                if (typeof data.Heartbeats != "object") {
                    list_lb.innerHTML = ''
                        + '<span class="' + css.nodata + '">'
                            + 'No Data '
                            + '(' + last_tracked + 's)'
                        + '</span>';

                    return false;
                }

                var html = '';
                for (const [id, utc]
                        of Object.entries(data.Heartbeats)) {

                    let ago = Math.round(Math.max((new Date()).getTime() - new Date(utc).getTime())/1000),
                        css_ago = (
                            (ago > 30) ? css.ago_danger :
                            (ago > 20) ? css.ago_warning :
                                         css.ago_nominal
                        );

                    html += ''
                    + '<span '
                        + 'title="' + id + '" '
                        + 'class="' + css.server + ' ' + css_ago + '">'
                        + ago + "s"
                    + '</span>';
                }
                list_lb.innerHTML = html;

                last_report = (new Date()).getTime();
            })
            .catch(err => {
                let last_tracked = Math.round((((new Date()).getTime() - last_report)/1000));

                onError(err, last_tracked);

                list_lb.innerHTML = ''
                    + '<span class="' + css.error + '">'
                        + 'Error: ' + err.message + ' '
                        + '(' + last_tracked + 's)'
                    + '</span>';
            });
    };

    list_lb.innerHTML = "loading...";
    setInterval(update, 1000);
    update();

    return {
        setAPIKey: setAPIKey,
        setOnData: setOnData,
        setOnError: setOnError,
    };
};
